from __future__ import print_function

import os
import sys
import re
import zipfile

from django.db import models
from django.db.models import Min, Max
from django.conf import settings
from django.utils.translation import ugettext, ugettext_lazy as _

try:
    from admin_steroids.utils import StringWithTitle
    APP_LABEL = StringWithTitle('django_sec', 'SEC')
except ImportError:
    APP_LABEL = 'django_sec'

from django_sec import xbrl

from . import constants as c
from .settings import DATA_DIR

def clean_unit_name(s):
    s = re.sub(r'[^a-z0-9]+', '', str(s).strip().lower())
    return s

class Namespace(models.Model):
    """
    Represents an XBRL namespace used to segment attribute names.
    """
    
    name = models.CharField(
        # Causes MySQL error:
        # Specified key was too long; max key length is 767 bytes
        # Note, PostreSQL has much higher limits.
        #max_length=255,
        max_length=100,
        blank=False,
        null=False,
        #db_index=True,
        unique=True)
    
    class Meta:
        app_label = APP_LABEL
    
    def __unicode__(self):
        return self.name

class UnitManager(models.Manager):
    
    def get_by_natural_key(self, name, *true_unit_nk):
        
        true_unit = None
        if true_unit_nk:
            true_unit = Unit.objects.get_by_natural_key(*true_unit_nk)
            
        u, _ = Unit.objects.get_or_create(name=name)
        u.true_unit = true_unit
        u.save()
        
        return u

class Unit(models.Model):
    """
    Represents a numeric unit.
    """
    
    objects = UnitManager()
    
    name = models.CharField(
        max_length=50,
        blank=False,
        null=False,
        db_index=True,
        unique=True)
    
    true_unit = models.ForeignKey(
        'self',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        help_text=_('''Points the the unit record this record duplicates.
            Points to itself if this is the master unit.'''))
    
    master = models.BooleanField(
        default=True,
        editable=False,
        help_text=_('If true, indicates this unit is the master referred to by duplicates.'))
    
    class Meta:
        app_label = APP_LABEL
        ordering = (
            'name',
        )
    
    def __unicode__(self):
        return self.name
    
    def natural_key(self):
        parts = (self.name,)
        if self.true_unit != self:
            parts += self.true_unit.natural_key()
        return parts 
        
    def save(self, *args, **kwargs):
        assert self.name.strip()
        if self.id:
            self.true_unit = self.true_unit or self
        self.master = self == self.true_unit
        if self.id:
            assert self.true_unit.master
        super(Unit, self).save(*args, **kwargs)
    
    @classmethod
    def do_update(cls, *args, **kwargs):
        q = cls.objects.filter(true_unit__isnull=True)
        for r in q.iterator():
            r.save()

class Attribute(models.Model):
    """
    Represents a financial attribute tag.
    """
    
    namespace = models.ForeignKey('Namespace')
    
    name = models.CharField(
        max_length=200,
        blank=False,
        null=False,
        db_index=True)
    
    load = models.BooleanField(
        default=False,
        db_index=True,
        help_text=_('If checked, all values will be loaded for this attribute.'))
    
    total_values = models.PositiveIntegerField(
        blank=True,
        null=True,
        editable=True)
    
    total_values_fresh = models.BooleanField(
        default=False,
        verbose_name='fresh')
    
    class Meta:
        app_label = APP_LABEL
        unique_together = (
            ('namespace', 'name'),
        )
        index_together = (
            ('namespace', 'name'),
        )
    
    def __unicode__(self):
        return '{%s}%s' % (self.namespace, self.name)
    
    @classmethod
    def do_update(cls, *args, **kwargs):
        q = cls.objects.filter(total_values_fresh=False).only('id', 'name')
        total = q.count()
        i = 0
        for r in q.iterator():
            i += 1
#            if not i % 100:
#                print('\rRefreshing attribute %i of %i.' % (i, total),
#                sys.stdout.flush()
            total_values = AttributeValue.objects.filter(attribute__name=r.name).count()
            cls.objects.filter(id=r.id).update(
                #total_values=r.values.all().count(),
                total_values=total_values,
                total_values_fresh=True)
#        print('\rRefreshing attribute %i of %i.' % (total, total),

class AttributeValue(models.Model):
    
    company = models.ForeignKey('Company', related_name='attributes')
    
    attribute = models.ForeignKey('Attribute', related_name='values')
    
    # Inspecting several XBRL samples, no digits above 12 characters
    # or decimals above 5 were found, so I've started there and added
    # a little more to handle future increases.
    value = models.DecimalField(
        max_digits=c.MAX_DIGITS,
        decimal_places=c.MAX_DECIMALS,
        blank=False,
        null=False)
    
    unit = models.ForeignKey('Unit')
    
    start_date = models.DateField(
        blank=False,
        null=False,
        db_index=True,
        help_text=_('''If attribute implies a duration, this is the date
            the duration begins. If the attribute implies an instance, this
            is the exact date it applies to.'''))
    
    end_date = models.DateField(
        blank=True,
        null=True,
        help_text=_('''If this attribute implies a duration, this is the date
            the duration ends.'''))
    
    filing_date = models.DateField(
        blank=False,
        null=False,
        help_text=_('The date this information became publically available.'))
    
    class Meta:
        app_label = APP_LABEL
        ordering = ('-attribute__total_values', '-start_date', 'attribute__name')
        unique_together = (
            ('company', 'attribute', 'start_date', 'end_date'),
        )
        index_together = (
            ('company', 'attribute', 'start_date'),
        )
        
    def __unicode__(self):
        return '%s %s=%s %s on %s' % (
            self.company,
            self.attribute.name,
            self.value,
            self.unit,
            self.start_date,
        )

class IndexFile(models.Model):
    
    year = models.IntegerField(
        blank=False,
        null=False,
        db_index=True)
    
    quarter = models.IntegerField(
        blank=False,
        null=False,
        db_index=True)
    
    filename = models.CharField(max_length=100, blank=False, null=False)
    
    total_rows = models.PositiveIntegerField(blank=True, null=True)
    
    processed_rows = models.PositiveIntegerField(blank=True, null=True)
    
    downloaded = models.DateTimeField(blank=True, null=True)
    
    processed = models.DateTimeField(blank=True, null=True)
    
    class Meta:
        app_label = APP_LABEL
        ordering = ('year', 'quarter')
        unique_together = (
            ('year', 'quarter'),
        )

class Company(models.Model):

    cik = models.IntegerField(
        db_index=True,
        primary_key=True,
        help_text=_('Central index key that uniquely identifies a filing entity.'))
    
    name = models.CharField(
        max_length=100,
        db_index=True,
        blank=False,
        null=False,
        help_text=_('The name of the company.'))
    
    load = models.BooleanField(
        default=False,
        db_index=True,
        help_text=_('If checked, all values for load-enabled attributes '
            'will be loaded for this company.'))
    
    min_date = models.DateField(
        blank=True,
        null=True,
        editable=False,
        db_index=True,
        help_text=_('''The oldest date of associated SEC Edgar filings
            for this company.'''))
    
    max_date = models.DateField(
        blank=True,
        null=True,
        editable=False,
        db_index=True,
        help_text=_('''The most recent date of associated SEC Edgar filings
            for this company.'''))
    
    class Meta:
        app_label = APP_LABEL
        verbose_name_plural = _('companies')
    
    def __unicode__(self):
        return self.name
    
    def save(self, *args, **kwargs):
        if self.cik:
            try:
                old = type(self).objects.get(cik=self.cik)
                
                aggs = self.attributes.all()\
                    .aggregate(Min('start_date'), Max('start_date'))
                self.min_date = aggs['start_date__min']
                self.max_date = aggs['start_date__max']
                
                if not old.load and self.load:
                    # If we just flag this company for loading then
                    # flag this company's indexes for loading.
                    Index.objects.filter(
                        company=self, attributes_loaded=True
                    ).update(attributes_loaded=False)
            except type(self).DoesNotExist:
                pass
        super(Company, self).save(*args, **kwargs)
    
class Index(models.Model):
    
    company = models.ForeignKey(
        'Company',
        related_name='filings')
    
    form = models.CharField(
        max_length=10,
        blank=True,
        db_index=True,
        verbose_name=_('form type'),
        help_text=_('The type of form the document is classified as.'))
    
    date = models.DateField(
        blank=False,
        null=False,
        db_index=True,
        verbose_name=_('date filed'),
        help_text=_('The date the item was filed with the SEC.'))
    
    filename = models.CharField(
        max_length=100,
        blank=False,
        null=False,
        db_index=True,
        help_text=_('The name of the associated financial filing.'))
    
    year = models.IntegerField(
        blank=False,
        null=False,
        db_index=True)
    
    quarter = models.IntegerField(
        blank=False,
        null=False,
        db_index=True)
    
    _ticker = models.CharField(
        max_length=50,
        db_index=True,
        db_column='ticker',
        verbose_name=_('ticker'),
        blank=True,
        null=True,
        help_text=_('''Caches the trading symbol if one is detected in the
            filing during attribute load.'''))
    
    attributes_loaded = models.BooleanField(default=False, db_index=True)
    
    valid = models.BooleanField(
        default=True,
        db_index=True,
        help_text=_('If false, errors were encountered trying to parse the associated files.'))
    
    error = models.TextField(blank=True, null=True)
    
    class Meta:
        app_label = APP_LABEL
        verbose_name_plural = _('indexes')
        unique_together = (
            # Note, filenames are not necessarily unique.
            # Filenames may be listed more than once under a different
            # form type.
            ('company', 'form', 'date', 'filename', 'year', 'quarter'),
        )
        index_together = (
            ('year', 'quarter'),
            ('company', 'date', 'filename'),
        )
        ordering = ('-date', 'filename')
    
    def xbrl_link(self):
        if self.form.startswith('10-K') or self.form.startswith('10-Q'):
            _id = self.filename.split('/')[-1][:-4]
            return 'http://www.sec.gov/Archives/edgar/data/%s/%s/%s-xbrl.zip' \
                % (self.company.cik, id.replace('-', ''), _id)
        return
        
    def html_link(self):
        return 'http://www.sec.gov/Archives/%s' % self.filename

    def index_link(self):
        _id = self.filename.split('/')[-1][:-4]
        return 'http://www.sec.gov/Archives/edgar/data/%s/%s/%s-index.htm' \
            % (self.company.cik, id.replace('-', ''), _id)
        
    def txt(self):
        return self.filename.split('/')[-1]
        
    def localfile(self):
        filename = '%s/%s/%s/%s' % (DATA_DIR, self.company.cik, self.txt()[:-4], self.txt())
        if os.path.exists(filename):
            return filename
        return
        
    def localpath(self):
        return '%s/%s/%s/' % (DATA_DIR, self.company.cik, self.txt()[:-4])

    def localcik(self):
        return '%s/%s/' % (DATA_DIR, self.company.cik)
    
    def html(self):
        filename = self.localfile()
        if not filename: 
            return
        f = open(filename, 'r').read()
        f_lower = f.lower()
        try:
            return f[f_lower.find('<html>'): f_lower.find('</html>')+4]
        except IndexError:
            print('html tag not found')
            return f

    def download(self, verbose=False):
        
        d = self.localcik()
        if not os.path.isdir(d):
            os.makedirs(d)
            
        d = self.localpath()
        if not os.path.isdir(d):
            os.makedirs(d)
            
        os.chdir(self.localpath())
        
        html_link = self.html_link()
        xbrl_link = self.xbrl_link()
        if verbose:
            print('xbrl_link:', xbrl_link)
            
#        if not os.path.exists(html_link.split('/')[-1]):
#            os.system('wget %s' % html_link)
        
        if xbrl_link:
            if not os.path.exists(xbrl_link.split('/')[-1]):
                if verbose:
                    os.system('wget %s' % xbrl_link)
                else:
                    os.system('wget --quiet %s' % xbrl_link)
                # Don't to this. It wastes disk space. Just read the ZIP directly.
                #os.system('unzip *.zip')

    def xbrl_localpath(self):
        try:
            os.chdir(self.localpath())
        except OSError:
            self.download()
        files = os.listdir('.')
#        print('files:',files
        archives = [elem for elem in files if elem.endswith('.zip')]
        if not archives:
            return None, None
        zf = zipfile.ZipFile(archives[0])
        #xml = sorted([elem for elem in files if elem.endswith('.xml')],key=len)
        xml = sorted([elem for elem in zf.namelist() if elem.endswith('.xml')], key=len)
#        print('xml:',xml
#        sys.exit()
        if not len(xml):
            return None, None
        #return self.localpath() + xml[0], zf.open
        return xml[0], zf.open

    def xbrl(self):
        filepath, open_method = self.xbrl_localpath()
#        print('filepath:',filepath
        if not filepath:
            print('no xbrl found. this option is for 10-ks.')
            return
        x = xbrl.XBRL(filepath, opener=open_method)
        x.fields['FiscalPeriod'] = x.fields['DocumentFiscalPeriodFocus']
        x.fields['FiscalYear'] = x.fields['DocumentFiscalYearFocus']
        x.fields['DocumentPeriodEndDate'] = x.fields['BalanceSheetDate']
        x.fields['PeriodStartDate'] = x.fields['IncomeStatementPeriodYTD']
        x.fields['SECFilingPage'] = self.index_link()
        x.fields['LinkToXBRLInstance'] = self.xbrl_link()
        return x
        
    def ticker(self):
        """
        Retrieves the company's stock ticker from an XML filing.
        Note, this is not guaranteed to exist.
        """
        if self._ticker:
            return self._ticker
        filepath, _ = self.xbrl_localpath()
        if filepath:
            ticker = filepath.split('/')[-1].split('-')[0].strip().upper()
            if ticker:
                self._ticker = ticker
            else:
                self._ticker = None
        return self._ticker



from django.forms import ModelForm

from . import models

class UnitChangeForm(ModelForm):
    
    class Meta:
        """Admin options."""
        model = models.Unit
        exclude = ()
    
    def __init__(self, *args, **kwargs):
        super(UnitChangeForm, self).__init__(*args, **kwargs)
        qs = models.Unit.objects.filter(master=True)
        self.fields["true_unit"].queryset = qs
        


import re
import time
from datetime import date

import dateutil.parser
import six
from six.moves import urllib as urllib2

try:
    from fake_useragent import UserAgent
except ImportError:
    UserAgent = None

def get_user_agent():
    if UserAgent:
        ua = UserAgent()
        return ua.random
    else:
        return 'Python-urllib/2.7/Django-SEC'

def str_to_date(s):
    s = str(s).strip()
    if not s:
        return
    dt = dateutil.parser.parse(s)
    if not dt:
        return
    return date(dt.year, dt.month, dt.day)

def lookup_cik(ticker, name=None):
    """
    Given a ticker symbol, retrieves the CIK.
    """
    ticker = ticker.strip().upper()
    
    # First try the SEC. In theory, should for all known symbols, even
    # deactivated ones. In practice, fails to work for many, even active ones.
    url = (
        'http://www.sec.gov/cgi-bin/browse-edgar?'
        'CIK={cik}&'
        'owner=exclude&'
        'Find=Find+Companies&'
        'action=getcompany'
    ).format(cik=ticker)
    request = urllib2.Request(url=url, headers={'User-agent':get_user_agent()})
    response = urllib2.urlopen(request)
    data = response.read()
    try:
        match = re.finditer(r'CIK=([0-9]+)', data).next()
        return match.group().split('=')[-1]
    except StopIteration:
        pass
    
    # Next, try SEC's other CIK lookup form.
    # It doesn't always work with just the ticker, so we also need to pass in
    # company name but it's the next most accurate after the first.
    # Unfortunately, this search is sensitive to punctuation in the company
    # name, which we might not have stored correctly.
    # So we start searching with everything we have, and then backoff to widen
    # the search.
    name = (name or '').strip()
    name = ''.join(_ for _ in (name or '').strip() if ord(_) < 128)
    if name:
        name_parts = name.split(' ')
        for i in six.moves.xrange(len(name_parts)):
            url = 'http://www.sec.gov/cgi-bin/cik.pl.c?company={company}'\
                .format(company='+'.join(name_parts[:-(i+1)]))
#            response = urllib2.urlopen(url)
            request = urllib2.Request(url=url, headers={'User-agent':get_user_agent()})
            response = urllib2.urlopen(request)
            data = response.read()
            matches = re.findall(r'CIK=([0-9]+)', data)
            if len(matches) == 1:
                return matches[0]
    
    # If the SEC search doesn't find anything, then try Yahoo.
    # Should work for all active symbols, but won't work for any deactive
    # symbols. 
    url = 'http://finance.yahoo.com/q/sec?s={symbol}+SEC+Filings'.format(symbol=ticker)
    #print('url2:',url
#    response = urllib2.urlopen(url)
    request = urllib2.Request(url=url, headers={'User-agent':get_user_agent()})
    response = urllib2.urlopen(request)
    data = response.read()
    try:
        match = re.finditer(r'search/\?cik=([0-9]+)', data).next()
        return match.group().split('=')[-1]
    except StopIteration:
        pass



import os
import sys

import django

PROJECT_DIR = os.path.dirname(__file__)

DATABASES = {
    'default':{
        'ENGINE': 'django.db.backends.sqlite3',
        # Don't do this. It dramatically slows down the test.
#        'NAME': '/tmp/django_sec.db',
#        'TEST_NAME': '/tmp/django_sec.db',
    }
}

ROOT_URLCONF = 'django_sec.tests.urls'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django_sec',
    'django_sec.tests',
]

MEDIA_ROOT = os.path.join(PROJECT_DIR, 'media')

# Disable migrations.
# http://stackoverflow.com/a/28560805/247542
class DisableMigrations(object):

    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return "notmigrations"
SOUTH_TESTS_MIGRATE = False # <= Django 1.8
#if django.VERSION > (1, 8, 0): # > Django 1.8
if django.VERSION > (1, 7, 0): # > Django 1.8 
    MIGRATION_MODULES = DisableMigrations()

USE_TZ = True

AUTH_USER_MODEL = 'auth.User'

SECRET_KEY = 'abc123'

SITE_ID = 1

BASE_SECURE_URL = 'https://localhost'

BASE_URL = 'http://localhost'

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    #'django.middleware.transaction.TransactionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.locale.LocaleMiddleware',    
)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-26 21:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=200)),
                ('load', models.BooleanField(db_index=True, default=False, help_text='If checked, all values will be loaded for this attribute.')),
                ('total_values', models.PositiveIntegerField(blank=True, null=True)),
                ('total_values_fresh', models.BooleanField(default=False, verbose_name=b'fresh')),
            ],
        ),
        migrations.CreateModel(
            name='AttributeValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.DecimalField(decimal_places=6, max_digits=40)),
                ('start_date', models.DateField(db_index=True, help_text='If attribute implies a duration, this is the date\n            the duration begins. If the attribute implies an instance, this\n            is the exact date it applies to.')),
                ('end_date', models.DateField(blank=True, help_text='If this attribute implies a duration, this is the date\n            the duration ends.', null=True)),
                ('filing_date', models.DateField(help_text='The date this information became publically available.')),
                ('attribute', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='values', to='django_sec.Attribute')),
            ],
            options={
                'ordering': ('-attribute__total_values', '-start_date', 'attribute__name'),
            },
        ),
        migrations.CreateModel(
            name='Company',
            fields=[
                ('cik', models.IntegerField(db_index=True, help_text='Central index key that uniquely identifies a filing entity.', primary_key=True, serialize=False)),
                ('name', models.CharField(db_index=True, help_text='The name of the company.', max_length=100)),
                ('load', models.BooleanField(db_index=True, default=False, help_text='If checked, all values for load-enabled attributes will be loaded for this company.')),
                ('min_date', models.DateField(blank=True, db_index=True, editable=False, help_text='The oldest date of associated SEC Edgar filings\n            for this company.', null=True)),
                ('max_date', models.DateField(blank=True, db_index=True, editable=False, help_text='The most recent date of associated SEC Edgar filings\n            for this company.', null=True)),
            ],
            options={
                'verbose_name_plural': 'companies',
            },
        ),
        migrations.CreateModel(
            name='Index',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('form', models.CharField(blank=True, db_index=True, help_text='The type of form the document is classified as.', max_length=10, verbose_name='form type')),
                ('date', models.DateField(db_index=True, help_text='The date the item was filed with the SEC.', verbose_name='date filed')),
                ('filename', models.CharField(db_index=True, help_text='The name of the associated financial filing.', max_length=100)),
                ('year', models.IntegerField(db_index=True)),
                ('quarter', models.IntegerField(db_index=True)),
                ('_ticker', models.CharField(blank=True, db_column=b'ticker', db_index=True, help_text='Caches the trading symbol if one is detected in the\n            filing during attribute load.', max_length=50, null=True, verbose_name='ticker')),
                ('attributes_loaded', models.BooleanField(db_index=True, default=False)),
                ('valid', models.BooleanField(db_index=True, default=True, help_text='If false, errors were encountered trying to parse the associated files.')),
                ('error', models.TextField(blank=True, null=True)),
                ('company', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='filings', to='django_sec.Company')),
            ],
            options={
                'ordering': ('-date', 'filename'),
                'verbose_name_plural': 'indexes',
            },
        ),
        migrations.CreateModel(
            name='IndexFile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('year', models.IntegerField(db_index=True)),
                ('quarter', models.IntegerField(db_index=True)),
                ('filename', models.CharField(max_length=100)),
                ('total_rows', models.PositiveIntegerField(blank=True, null=True)),
                ('processed_rows', models.PositiveIntegerField(blank=True, null=True)),
                ('downloaded', models.DateTimeField(blank=True, null=True)),
                ('processed', models.DateTimeField(blank=True, null=True)),
            ],
            options={
                'ordering': ('year', 'quarter'),
            },
        ),
        migrations.CreateModel(
            name='Namespace',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name='Unit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=50, unique=True)),
                ('master', models.BooleanField(default=True, editable=False, help_text='If true, indicates this unit is the master referred to by duplicates.')),
                ('true_unit', models.ForeignKey(blank=True, help_text='Points the the unit record this record duplicates.\n            Points to itself if this is the master unit.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='django_sec.Unit')),
            ],
            options={
                'ordering': ('name',),
            },
        ),
        migrations.AlterUniqueTogether(
            name='indexfile',
            unique_together=set([('year', 'quarter')]),
        ),
        migrations.AddField(
            model_name='attributevalue',
            name='company',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attributes', to='django_sec.Company'),
        ),
        migrations.AddField(
            model_name='attributevalue',
            name='unit',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_sec.Unit'),
        ),
        migrations.AddField(
            model_name='attribute',
            name='namespace',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_sec.Namespace'),
        ),
        migrations.AlterUniqueTogether(
            name='index',
            unique_together=set([('company', 'form', 'date', 'filename', 'year', 'quarter')]),
        ),
        migrations.AlterIndexTogether(
            name='index',
            index_together=set([('company', 'date', 'filename'), ('year', 'quarter')]),
        ),
        migrations.AlterUniqueTogether(
            name='attributevalue',
            unique_together=set([('company', 'attribute', 'start_date', 'end_date')]),
        ),
        migrations.AlterIndexTogether(
            name='attributevalue',
            index_together=set([('company', 'attribute', 'start_date')]),
        ),
        migrations.AlterUniqueTogether(
            name='attribute',
            unique_together=set([('namespace', 'name')]),
        ),
        migrations.AlterIndexTogether(
            name='attribute',
            index_together=set([('namespace', 'name')]),
        ),
    ]



from __future__ import print_function

import urllib
import os
import re
import sys
from zipfile import ZipFile
import time
from datetime import date, datetime, timedelta
from optparse import make_option
import traceback
import random
from multiprocessing import Process, Lock, Queue
import collections

import django
from django.core.management.base import BaseCommand
from django.db import transaction, connection, IntegrityError, DatabaseError
from django.db.models import Q, F
from django.conf import settings
from django.utils import timezone

from django_sec import models
from django_sec.models import DATA_DIR, c

try:
    from psycopg2.extensions import TransactionRollbackError
except ImportError:
    TransactionRollbackError = Exception

try:
    from chroniker.models import Job
except ImportError:
    Job = None

def get_options(parser=None):
    make_opt = make_option
    if parser:
        make_opt = parser.add_argument
    return [
        make_opt('--name', default=None),
    ]

class Command(BaseCommand):
    help = "Links duplicate units to the true canonical unit."
    args = ''
    option_list = getattr(BaseCommand, 'option_list', ()) + tuple(get_options())
    
    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            get_options(parser)
            self.add_arguments(parser)
        return parser
    
    def handle(self, **options):
        
        settings.DEBUG = False
        
        only_name = options['name']
        
        qs = models.Unit.objects.all()
        if only_name:
            qs = qs.filter(name__icontains=only_name)
        total = qs.count()
        i = 0
        dups = set()

        # Link all singular to plurals.
        qs = models.Unit.objects.filter(master=True)
        total = qs.count()
        i = 0
        for r in qs.iterator():
            i += 1
            sys.stdout.write('\r%i of %i' % (i, total))
            sys.stdout.flush()
            
            plural_qs = models.Unit.objects\
                .filter(master=True)\
                .filter(Q(name=r.name+'s')|Q(name=r.name+'es'))\
                .exclude(id=r.id)
            if plural_qs.exists():
                models.Unit.objects.filter(true_unit=r)\
                    .update(true_unit=plural_qs[0], master=False)
                r.true_unit = plural_qs[0]
                r.master = False
                r.save()

        print()       
        print('%i duplicates linked' % len(dups))



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding field 'Company.min_date'
        db.add_column(u'django_sec_company', 'min_date',
                      self.gf('django.db.models.fields.DateField')(db_index=True, null=True, blank=True),
                      keep_default=False)

        # Adding field 'Company.max_date'
        db.add_column(u'django_sec_company', 'max_date',
                      self.gf('django.db.models.fields.DateField')(db_index=True, null=True, blank=True),
                      keep_default=False)


    def backwards(self, orm):
        # Deleting field 'Company.min_date'
        db.delete_column(u'django_sec_company', 'min_date')

        # Deleting field 'Company.max_date'
        db.delete_column(u'django_sec_company', 'max_date')


    models = {
        'django_sec.attribute': {
            'Meta': {'unique_together': "(('namespace', 'name'),)", 'object_name': 'Attribute', 'index_together': "(('namespace', 'name'),)"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'load': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '500', 'db_index': 'True'}),
            'namespace': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['django_sec.Namespace']"}),
            'total_values': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'total_values_fresh': ('django.db.models.fields.BooleanField', [], {'default': 'False'})
        },
        'django_sec.attributevalue': {
            'Meta': {'ordering': "('-attribute__total_values', '-start_date', 'attribute__name')", 'unique_together': "(('company', 'attribute', 'start_date', 'end_date'),)", 'object_name': 'AttributeValue', 'index_together': "(('company', 'attribute', 'start_date'),)"},
            'attribute': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'values'", 'to': "orm['django_sec.Attribute']"}),
            'company': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'attributes'", 'to': "orm['django_sec.Company']"}),
            'end_date': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'filing_date': ('django.db.models.fields.DateField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'start_date': ('django.db.models.fields.DateField', [], {'db_index': 'True'}),
            'unit': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['django_sec.Unit']"}),
            'value': ('django.db.models.fields.DecimalField', [], {'max_digits': '40', 'decimal_places': '6'})
        },
        'django_sec.company': {
            'Meta': {'object_name': 'Company'},
            'cik': ('django.db.models.fields.IntegerField', [], {'primary_key': 'True', 'db_index': 'True'}),
            'load': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'max_date': ('django.db.models.fields.DateField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'min_date': ('django.db.models.fields.DateField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'db_index': 'True'})
        },
        'django_sec.index': {
            'Meta': {'ordering': "('-date', 'filename')", 'unique_together': "(('company', 'form', 'date', 'filename', 'year', 'quarter'),)", 'object_name': 'Index', 'index_together': "(('year', 'quarter'), ('company', 'date', 'filename'))"},
            '_ticker': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '50', 'null': 'True', 'db_column': "'ticker'", 'blank': 'True'}),
            'attributes_loaded': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'company': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'filings'", 'to': "orm['django_sec.Company']"}),
            'date': ('django.db.models.fields.DateField', [], {'db_index': 'True'}),
            'error': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'filename': ('django.db.models.fields.CharField', [], {'max_length': '100', 'db_index': 'True'}),
            'form': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '10', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'quarter': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'}),
            'valid': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'year': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'})
        },
        'django_sec.indexfile': {
            'Meta': {'ordering': "('year', 'quarter')", 'unique_together': "(('year', 'quarter'),)", 'object_name': 'IndexFile'},
            'downloaded': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'filename': ('django.db.models.fields.CharField', [], {'max_length': '200'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'processed': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'processed_rows': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'quarter': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'}),
            'total_rows': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'year': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'})
        },
        'django_sec.namespace': {
            'Meta': {'object_name': 'Namespace'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '500', 'db_index': 'True'})
        },
        'django_sec.unit': {
            'Meta': {'ordering': "('name',)", 'object_name': 'Unit'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'master': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '200', 'db_index': 'True'}),
            'true_unit': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['django_sec.Unit']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'})
        }
    }

    complete_apps = ['django_sec']


from datetime import date
import re

from lxml import etree
import dateutil.parser

from .xbrl_fundamentals import FundamentantalAccountingConcepts
from . import constants as c
from . import utils

class XBRL:

    def __init__(self, XBRLInstanceLocation, opener=None):
        self.XBRLInstanceLocation = XBRLInstanceLocation
        self.fields = {}
        
        if opener:
            # Allow us to read directly from a ZIP archive without extracting
            # the whole thing.
            self.EntireInstanceDocument = opener(XBRLInstanceLocation, 'r').read()
        else:
            self.EntireInstanceDocument = open(XBRLInstanceLocation, 'r').read()
         
        self.oInstance = etree.fromstring(self.EntireInstanceDocument)
        self.ns = {}
        for k in self.oInstance.nsmap.keys():
            if k != None:
                self.ns[k] = self.oInstance.nsmap[k]
        self.ns['xbrli'] = 'http://www.xbrl.org/2003/instance'
        self.ns['xlmns'] = 'http://www.xbrl.org/2003/instance'
        self.GetBaseInformation()
        self.loadYear(0)
        
        self._context_start_dates = {}
        self._context_end_dates = {}

    def loadYear(self, yearminus=0):
        currentEnd = self.getNode("//dei:DocumentPeriodEndDate").text
        asdate = re.match(r'\s*(\d{4})-(\d{2})-(\d{2})\s*', currentEnd)
        if asdate:
            year = int(asdate.groups()[0]) - yearminus
            thisend = '%s-%s-%s' % (year, asdate.groups()[1], asdate.groups()[2])
            self.GetCurrentPeriodAndContextInformation(thisend)
            FundamentantalAccountingConcepts(self)
            return True
        else:
            print(currentEnd, ' is not a date')
            return False
            
    def getNodeList(self, xpath, root=None):
        if root is None:
            root = self.oInstance
        oNodelist = root.xpath(xpath, namespaces=self.ns)
        return oNodelist
        
    def getNode(self, xpath, root=None):
        oNodelist = self.getNodeList(xpath, root)
        if len(oNodelist):
            return oNodelist[0]
        return None

    def iter_namespace(self, ns='us-gaap'):
        """
        Iterates over all namespace elements, yielding each one.
        """
        SeekConcept = '%s:*' % (ns,)
        node_list = self.getNodeList("//" + SeekConcept)
        total = len(node_list)
        for node in node_list:
            yield node, total

    def GetFactValue(self, SeekConcept, ConceptPeriodType):
                
        factValue = None
            
        if ConceptPeriodType == c.INSTANT:
            ContextReference = self.fields['ContextForInstants']
        elif ConceptPeriodType == c.DURATION:
            ContextReference = self.fields['ContextForDurations']
        else:
            #An error occured
            return "CONTEXT ERROR"
        
        if not ContextReference:
            return None

        oNode = self.getNode("//" + SeekConcept + "[@contextRef='" + ContextReference + "']")
        if oNode is not None:
            factValue = oNode.text
            if 'nil' in oNode.keys() and oNode.get('nil') == 'true':
                factValue = 0
                #set the value to ZERO if it is nil
            #if type(factValue)==str:
            try:
                factValue = float(factValue)
            except (ValueError, TypeError):
                #print('couldnt convert %s=%s to string' % (SeekConcept,factValue)
                factValue = None
            
        return factValue

    def GetBaseInformation(self):
                
        #Registered Name
        oNode = self.getNode("//dei:EntityRegistrantName[@contextRef]")
        if oNode is not None:
            self.fields['EntityRegistrantName'] = oNode.text
        else:
            self.fields['EntityRegistrantName'] = "Registered name not found"

        #Fiscal year
        oNode = self.getNode("//dei:CurrentFiscalYearEndDate[@contextRef]")        
        if oNode is not None:
            self.fields['FiscalYear'] = oNode.text
        else:
            self.fields['FiscalYear'] = "Fiscal year not found"

        #EntityCentralIndexKey
        oNode = self.getNode("//dei:EntityCentralIndexKey[@contextRef]")
        if oNode is not None:
            self.fields['EntityCentralIndexKey'] = oNode.text
        else:
            self.fields['EntityCentralIndexKey'] = "CIK not found"

        #EntityFilerCategory
        oNode = self.getNode("//dei:EntityFilerCategory[@contextRef]")
        if oNode is not None:
            self.fields['EntityFilerCategory'] = oNode.text
        else:
            self.fields['EntityFilerCategory'] = "Filer category not found"

        #TradingSymbol
        oNode = self.getNode("//dei:TradingSymbol[@contextRef]")
        if oNode is not None:
            self.fields['TradingSymbol'] = oNode.text
        else:
            self.fields['TradingSymbol'] = None

        #DocumentFiscalYearFocus
        oNode = self.getNode("//dei:DocumentFiscalYearFocus[@contextRef]")
        if oNode is not None:
            self.fields['DocumentFiscalYearFocus'] = oNode.text
        else:
            self.fields['DocumentFiscalYearFocus'] = "Fiscal year focus not found"

        #DocumentFiscalPeriodFocus
        oNode = self.getNode("//dei:DocumentFiscalPeriodFocus[@contextRef]")
        if oNode is not None:
            self.fields['DocumentFiscalPeriodFocus'] = oNode.text
        else:
            self.fields['DocumentFiscalPeriodFocus'] = "Fiscal period focus not found"
        
        #DocumentType
        oNode = self.getNode("//dei:DocumentType[@contextRef]")
        if oNode is not None:
            self.fields['DocumentType'] = oNode.text
        else:
            self.fields['DocumentType'] = "Fiscal period focus not found"
        
    def get_context_start_date(self, context_id):
        if context_id not in self._context_start_dates:
            node = self.getNode(
                "//xbrli:context[@id='" + context_id + "']/xbrli:period/xbrli:startDate")
            if node is None:
                node = self.getNode(
                    "//xbrli:context[@id='" + context_id + "']/xbrli:period/xbrli:instant")
            dt = None
            if node is not None and node.text:
                #dt = date(*map(int, node.text.split('-')))
                dt = utils.str_to_date(node.text)
            self._context_start_dates[context_id] = dt
        return self._context_start_dates[context_id]

    def get_context_end_date(self, context_id):
        if context_id not in self._context_end_dates:
            node = self.getNode(
                "//xbrli:context[@id='" + context_id + "']/xbrli:period/xbrli:endDate")
            dt = None
            if node is not None and node.text:
                #dt = date(*map(int, node.text.split('-')))
                dt = utils.str_to_date(node.text)
            self._context_end_dates[context_id] = dt
        return self._context_end_dates[context_id]
        
    def GetCurrentPeriodAndContextInformation(self, EndDate):
        # Figures out the current period and contexts for the current
        # period instance/duration contexts

        self.fields['BalanceSheetDate'] = "ERROR"
        self.fields['IncomeStatementPeriodYTD'] = "ERROR"
        
        self.fields['ContextForInstants'] = "ERROR"
        self.fields['ContextForDurations'] = "ERROR"

        # This finds the period end date for the database table,
        # and instant date (for balance sheet):        
        UseContext = "ERROR"
        #EndDate = self.getNode("//dei:DocumentPeriodEndDate").text
        #This is the <instant> or the <endDate>
        
        # Uses the concept ASSETS to find the correct instance context
        # This finds the Context ID for that end date (has correct <instant>
        # date plus has no dimensions):    
        oNodelist2 = self.getNodeList(
            "//us-gaap:Assets | "
            "//us-gaap:AssetsCurrent | "
            "//us-gaap:LiabilitiesAndStockholdersEquity")
             
        ContextForInstants = UseContext
        self.fields['ContextForInstants'] = ContextForInstants
        
        ###This finds the duration context
        ###This may work incorrectly for fiscal year ends because the dates cross calendar years
        #Get context ID of durations and the start date for the database table
        oNodelist2 = self.getNodeList(
            "//us-gaap:CashAndCashEquivalentsPeriodIncreaseDecrease | "
            "//us-gaap:CashPeriodIncreaseDecrease | "
            "//us-gaap:NetIncomeLoss | "
            "//dei:DocumentPeriodEndDate")

        StartDate = "ERROR"
        StartDateYTD = "2099-01-01"
        UseContext = "ERROR"
        
        #Balance sheet date of current period
        self.fields['BalanceSheetDate'] = EndDate
        
        #MsgBox "Instant context is: " + ContextForInstants
        if ContextForInstants == "ERROR":
            #MsgBox "Looking for alternative instance context"
            
            ContextForInstants = self.LookForAlternativeInstanceContext()
            self.fields['ContextForInstants'] = ContextForInstants
        
        
        #Income statement date for current fiscal year, year to date
        self.fields['IncomeStatementPeriodYTD'] = StartDateYTD
        
        ContextForDurations = UseContext
        self.fields['ContextForDurations'] = ContextForDurations

    def LookForAlternativeInstanceContext(self):
        #This deals with the situation where no instance context has no dimensions
        #Finds something
            
        something = None
        
        #See if there are any nodes with the document period focus date
        oNodeList_Alt = self.getNodeList(
            "//xbrli:context[xbrli:period/xbrli:instant='" + \
            self.fields['BalanceSheetDate'] + "']")

        #MsgBox "Node list length: " + oNodeList_Alt.length
        for oNode_Alt in oNodeList_Alt:
            #Found possible contexts
            #MsgBox oNode_Alt.selectSingleNode("@id").text
            something = self.getNode("//us-gaap:Assets[@contextRef='" + oNode_Alt.get("id") + "']")
            if something is not None:
                #MsgBox "Use this context: " + oNode_Alt.selectSingleNode("@id").text
                return oNode_Alt.get("id")



from __future__ import print_function

import urllib
import os
import re
import sys
from zipfile import ZipFile
import time
from datetime import date, datetime, timedelta
from optparse import make_option
import traceback
import random
from multiprocessing import Process, Lock, Queue
import collections

import six
from six import StringIO, string_types

import django
from django.core.management.base import BaseCommand
from django.db import transaction, connection, IntegrityError, DatabaseError
from django.conf import settings
from django.utils import timezone

from django_sec import models
from django_sec.models import DATA_DIR, c

try:
    from psycopg2.extensions import TransactionRollbackError
except ImportError:
    TransactionRollbackError = Exception

try:
    from chroniker.models import Job
except ImportError:
    Job = None

def is_power_of_two(x):
    return (x & (x - 1)) == 0

def parse_stripe(stripe):
    stripe_num = None
    stripe_mod = None
    if stripe:
        assert isinstance(stripe, string_types) and len(stripe) == 2
        stripe_num, stripe_mod = stripe
        stripe_num = int(stripe_num)
        stripe_mod = int(stripe_mod)
        assert stripe_num < stripe_mod
    return stripe_num, stripe_mod

def get_options(parser=None):
    make_opt = make_option
    if parser:
        make_opt = parser.add_argument
    return [
        make_opt('--cik',
            default=None),
        make_opt('--forms',
            default='10-K,10-Q'),
        make_opt('--start-year',
            default=None),
        make_opt('--end-year',
            default=None),
        make_opt('--quarter',
            default=None),
#         make_opt('--dryrun',
#             action='store_true',
#             default=False),
        make_opt('--force',
            action='store_true',
            default=False),
        make_opt('--verbose',
            action='store_true',
            default=False),
        make_opt('--multi',
            dest='multi',
            default=0,
            help='The number of processes to use. Must be a multiple of 2.'),
        make_opt('--show-pending',
            action='store_true',
            default=False,
            help='If given, will only report the number of pending records to process then exit.'),
    ]

class Command(BaseCommand):
    help = "Shows data from filings."
    args = ''
    option_list = getattr(BaseCommand, 'option_list', ()) + tuple(get_options())
        
    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            get_options(parser)                
            self.add_arguments(parser)
        return parser
    
    def handle(self, **options):
        
#         self.dryrun = options['dryrun']
        self.force = options['force']
        self.verbose = options['verbose']
        
        self.stripe_counts = {} # {stripe:{current,total}
        self.last_progress_refresh = None
        self.start_times = {} # {key:start_time}
        
        self.cik = (options['cik'] or '').strip()
        if self.cik:
            self.cik = int(self.cik)
        else:
            self.cik = None
        
        self.forms = (options['forms'] or '').strip().split(',')
        
        start_year = options['start_year']
        if start_year:
            start_year = int(start_year)
        else:
            #start_year = date.today().year - 1
            start_year = 1900
        self.start_year = start_year
            
        end_year = options['end_year']
        if end_year:
            end_year = int(end_year)
        else:
            end_year = date.today().year
        self.end_year = end_year

        self.status = None
        self.progress = collections.OrderedDict()
        multi = int(options['multi'])
        kwargs = options.copy()
        if multi:
            assert multi > 1 and is_power_of_two(multi), \
                "Process count must be greater than 1 and a multiple of 2."
            processes = []
            self.status = Queue()
            for i, _ in enumerate(six.moves.range(multi)):
                print('Starting process %i' % i)
                stripe = kwargs['stripe'] = '%i%i' % (i, multi)
                kwargs['status'] = self.status
                
                connection.close()
                p = Process(target=self.run_process, kwargs=kwargs)
                p.daemon = True
                processes.append(p)
                p.start()
            self.progress[stripe] = (0, 0, 0, 0, None, '')
            #return
            while any(i.is_alive() for i in processes):
                time.sleep(0.1)
                while not self.status.empty():
                    stripe, current, total, sub_current, sub_total, eta, message = \
                        self.status.get()
                    self.progress[stripe] = (current, total, sub_current, sub_total, eta, message)
                    if stripe not in self.start_times:
                        self.start_times[stripe] = time.time()
                    self.print_progress()
            print('All processes complete.')
        else:
            self.start_times[None] = time.time()
            self.run_process(**kwargs)
    
    def print_progress(self, clear=True, newline=True):
        td = datetime.now() - self.last_progress_refresh
        if self.last_progress_refresh and td.seconds < 0.5:
            return
        bar_length = 10
        if clear:
            sys.stdout.write('\033[2J\033[H') #clear screen
            sys.stdout.write('Importing attributes\n')
        for stripe, msg_parts in sorted(self.progress.items()):
            (current, total, sub_current, sub_total, eta, message) = msg_parts
            sub_status = ''
            if total:
                if not eta:
                    start_time = self.start_times[stripe]
                    current_seconds = time.time() - start_time
                    total_seconds = float(total)/current*current_seconds
                    remaining_seconds = int(total_seconds - current_seconds)
                    eta = timezone.now() + timedelta(seconds=remaining_seconds)
                    
                self.stripe_counts[stripe] = (current, total)
                percent = current/float(total)
                bar = ('=' * int(percent * bar_length)).ljust(bar_length)
                percent = int(percent * 100)
            else:
                eta = eta or '?'
                percent = 0
                bar = ('=' * int(percent * bar_length)).ljust(bar_length)
                percent = '?'
                total = '?'
            if sub_current and sub_total:
                sub_status = '(subtask %s of %s) ' % (sub_current, sub_total)
            sys.stdout.write(
                (('' if newline else '\r') + \
                "%s [%s] %s of %s %s%s%% eta=%s: %s"+('\n' if newline else '')) \
                    % (stripe, bar, current, total, sub_status, percent, eta, message))
        sys.stdout.flush()
        self.last_progress_refresh = datetime.now()
        
        # Update job.
        overall_current_count = 0
        overall_total_count = 0
        for stripe, (current, total) in self.stripe_counts.iteritems():
            overall_current_count += current
            overall_total_count += total
        #print('overall_current_count:',overall_current_count
        #print('overall_total_count:',overall_total_count
        if overall_total_count and Job:
            Job.update_progress(
                total_parts_complete=overall_current_count,
                total_parts=overall_total_count,
            )
#             if not self.dryrun:
#                 transaction.commit()
    
    def run_process(self, status=None, **kwargs):
        tmp_debug = settings.DEBUG
        settings.DEBUG = False
        #transaction.enter_transaction_management()
        #transaction.managed(True)
        try:
            print('Running process:', kwargs)
            self.import_attributes(status=status, **kwargs)
            print('Done process:', kwargs)
        finally:
            settings.DEBUG = tmp_debug
            #if self.dryrun:
            #    print('This is a dryrun, so no changes were committed.'
            #    transaction.rollback()
            #else:
            #    transaction.commit()
            #transaction.leave_transaction_management()
            connection.close()
    
    def import_attributes(self, status=None, **kwargs):
        stripe = kwargs.get('stripe')
        reraise = kwargs.get('reraise')
        
        current_count = 0
        total_count = 0
        fatal_errors = False
        fatal_error = None
        estimated_completion_datetime = None
        sub_current = 0
        sub_total = 0
        
        def print_status(message, count=None, total=None):
            #print('message:',message
            current_count = count or 0
            total_count = total or 0
            if status:
                status.put([
                    stripe,
                    current_count+1,
                    total_count,
                    sub_current,
                    sub_total,
                    estimated_completion_datetime,
                    message,
                ])
            else:
                #print('total_count:',total_count
                self.progress[stripe] = (
                    current_count,
                    total_count,
                    sub_current,
                    sub_total,
                    estimated_completion_datetime,
                    message,
                )
                self.print_progress(clear=False, newline=True)
        
        stripe_num, stripe_mod = parse_stripe(stripe)
        if stripe:
            print_status('Striping with number %i and modulus %i.' \
                % (stripe_num, stripe_mod))
        
        try:
            # Get a file from the index.
            # It may or may not be present on our hard disk.
            # If it's not, it will be downloaded
            # the first time we try to access it, or you can call
            # .download() explicitly.
            q = models.Index.objects.filter(
                year__gte=self.start_year,
                year__lte=self.end_year)
            if not self.force:
                q = q.filter(
                    attributes_loaded__exact=0,#False,
                    valid__exact=1,#True,
                )
            if self.forms:
                q = q.filter(form__in=self.forms)
            
            q2 = q
            if self.cik:
                q = q.filter(company__cik=self.cik, company__load=True)
                q2 = q2.filter(company__cik=self.cik)
                if not q.count() and q2.count():
                    print(('Warning: the company you specified with cik %s is not '
                        'marked for loading.') % (self.cik,), file=sys.stderr)
            
            if stripe is not None:
                q = q.extra(
                    where=['((django_sec_index.id %%%% %i) = %i)' % (stripe_mod, stripe_num)])
                    
            #print_status('Finding total record count...')
            #print('query:', q.query
            total_count = total = q.count()
            print('total_count:', total_count)
            
            if kwargs['show_pending']:
                print('='*80)
                print('%i total pending records' % total_count)
                return
            
            print_status('%i total rows.' % (total,))
            i = 0
            commit_freq = 100
            print_status('%i indexes found for forms %s.' \
                % (total, ', '.join(self.forms)), count=0, total=total)
            for ifile in q.iterator():
                i += 1
                current_count = i
                
                msg = 'Processing index %s.' % (ifile.filename,)
                print_status(msg, count=i, total=total)
                
                if not i % commit_freq:
                    sys.stdout.flush()
                
                ifile.download(verbose=self.verbose)
                
                # Initialize XBRL parser and populate an attribute called fields with
                # a dict of 50 common terms.
                x = None
                error = None
                try:
                    x = ifile.xbrl()
                except Exception as e:
                    ferr = StringIO()
                    traceback.print_exc(file=ferr)
                    error = ferr.getvalue()
                
                if x is None:
                    if error is None:
                        error = 'No XBRL found.'
                    models.Index.objects.filter(id=ifile.id)\
                        .update(valid=False, error=error)
                    continue
                
                maxretries = 10
                retry = 0
                #for retry in xrange(maxretries):
                while 1:
                    try:
                        
                        #x.loadYear(2)
            #            print'Year:', x.fields['FiscalYear']
                        company = ifile.company
                        max_text_len = 0
                        unique_attrs = set()
                        bulk_objects = []
                        prior_keys = set()
                        j = sub_total = 0
                        #print
                        for node, sub_total in x.iter_namespace():
                            j += 1
                            sub_current = j
                            if not j % commit_freq:
                                print_status(msg, count=i, total=total)
                                #sys.stdout.flush()
#                                 if not self.dryrun:
#                                     transaction.commit()
                                
                            matches = re.findall(r'^\{([^\}]+)\}(.*)$', node.tag)
                            if matches:
                                ns, attr_name = matches[0]
                            else:
                                ns = None
                                attr_name = node
                            decimals = node.attrib.get('decimals', None)
                            if decimals is None:
                                continue
                            if decimals.upper() == 'INF':
                                decimals = 6
                            decimals = int(decimals)
                            max_text_len = max(max_text_len, len((node.text or '').strip()))
                            context_id = node.attrib['contextRef']
                #            if context_id != 'D2009Q4YTD':
                #                continue
                            start_date = x.get_context_start_date(context_id)
                            if not start_date:
                                continue
                            end_date = x.get_context_end_date(context_id)
                            if not end_date:
                                continue
                            namespace, _ = models.Namespace.objects.get_or_create(name=ns.strip())
                            attribute, _ = models.Attribute.objects.get_or_create(
                                namespace=namespace,
                                name=attr_name,
                                defaults=dict(load=True),
                            )
                            if not attribute.load:
                                continue
                            unit, _ = models.Unit.objects.get_or_create(
                                name=node.attrib['unitRef'].strip())
                            unit.save()
                            value = (node.text or '').strip()
                            if not value:
                                continue
                            assert len(value.split('.')[0]) <= c.MAX_QUANTIZE, \
                                'Value too large, must be less than %i digits: %i %s' \
                                    % (c.MAX_QUANTIZE, len(value), repr(value))
                            
                            #print(attribute
                            models.Attribute.objects\
                                .filter(id=attribute.id)\
                                .update(total_values_fresh=False)
                            
                            if models.AttributeValue.objects\
                            .filter(company=company, attribute=attribute, start_date=start_date)\
                            .exists():
                                continue
            
                            # Some attributes are listed multiple times in differently
                            # named contexts even though the value and date ranges are
                            # identical.
                            key = (company, attribute, start_date)
                            if key in prior_keys:
                                continue
                            prior_keys.add(key)
                            
                            bulk_objects.append(models.AttributeValue(
                                company=company,
                                attribute=attribute,
                                start_date=start_date,
                                end_date=end_date,
                                value=value,
                                unit=unit,
                                filing_date=ifile.date,
                            ))
                            
                            if not len(bulk_objects) % commit_freq:
                                models.AttributeValue.objects.bulk_create(bulk_objects)
                                bulk_objects = []
                                prior_keys.clear()
                                
#                         if not self.dryrun:
#                             transaction.commit()
            #            print('\rImporting attribute %i of %i.' % (sub_total, sub_total),
            #            print
                        print_status('Importing attributes.', count=i, total=total)
                        
                        if bulk_objects:
                            models.AttributeValue.objects.bulk_create(bulk_objects)
                            bulk_objects = []
                            
                        ticker = ifile.ticker()
                        models.Index.objects\
                            .filter(id=ifile.id)\
                            .update(attributes_loaded=True, _ticker=ticker)
                        
                        models.Attribute.do_update()
                        
                        models.Unit.do_update()
                        
#                         if not self.dryrun:
#                             transaction.commit()
                    
                        break
                    
                    except DatabaseError as e:
                        if retry+1 == maxretries:
                            raise
                        print(e, 'retry', retry)
                        connection.close()
                        time.sleep(random.random()*5)
                    
                    except TransactionRollbackError as e:
                        if TransactionRollbackError.__name__ != 'TransactionRollbackError':
                            raise
                        if retry+1 == maxretries:
                            raise
                        print(e, 'retry', retry)
                        connection.close()
                        time.sleep(random.random()*5)
                    
        except Exception as e:
            print('Error: %s' % e)
            ferr = StringIO()
            traceback.print_exc(file=ferr)
            error = ferr.getvalue()
            print_status('Fatal error: %s' % (error,))
        finally:
            connection.close()



from __future__ import print_function

import urllib
import os
import re
import sys
from zipfile import ZipFile
import time
from datetime import date, datetime, timedelta
from optparse import make_option

import django
from django.core.management.base import BaseCommand
from django.db import transaction, connection
from django.conf import settings
from django.utils import timezone
from django.utils.encoding import force_text

from django_sec.models import Company, Index, IndexFile, DATA_DIR

def removeNonAscii(s):
    return "".join(i for i in s if ord(i) < 128)

def get_options(parser=None):
    make_opt = make_option
    if parser:
        make_opt = parser.add_argument
    return [
        make_opt('--start-year',
            default=None),
        make_opt('--end-year',
            default=None),
        make_opt('--quarter',
            default=None),
        make_opt('--delete-prior-indexes',
            action='store_true',
            default=False),
        make_opt('--reprocess',
            action='store_true',
            default=False),
        make_opt('--auto-reprocess-last-n-days',
            default=90,
            help='The number of days to automatically redownload and reprocess index files.'),
    ]

class Command(BaseCommand):
    help = "Download new files representing one month of 990s, ignoring months we already have. "\
        "Each quarter contains hundreds of thousands of filings; will take a while to run. "
    #args = ''
    option_list = getattr(BaseCommand, 'option_list', ()) + tuple(get_options())
        
    def create_parser(self, prog_name, subcommand):
        """
        For ``Django>=1.10``
        Create and return the ``ArgumentParser`` which extends ``BaseCommand`` parser with
        chroniker extra args and will be used to parse the arguments to this command.
        """
        from distutils.version import StrictVersion # pylint: disable=E0611
        parser = super(Command, self).create_parser(prog_name, subcommand)
        version_threshold = StrictVersion('1.10')
        current_version = StrictVersion(django.get_version(django.VERSION))
        if current_version >= version_threshold:
            get_options(parser)
            self.add_arguments(parser)
        return parser
    
    def handle(self, **options):
        
        start_year = options['start_year']
        if start_year:
            start_year = int(start_year)
        else:
            start_year = date.today().year - 1
            
        end_year = options['end_year']
        if end_year:
            end_year = int(end_year)
        else:
            end_year = date.today().year+1
        
        reprocess = options['reprocess']
        
        target_quarter = options['quarter']
        if target_quarter:
            target_quarter = int(target_quarter)
        
        auto_reprocess_last_n_days = int(options['auto_reprocess_last_n_days'])
        
        tmp_debug = settings.DEBUG
        settings.DEBUG = False
        try:
            for year in range(start_year, end_year):
                for quarter in range(4):
                    if target_quarter and quarter+1 != target_quarter:
                        continue
                    quarter_start = date(year, quarter*3+1, 1)
                    cutoff_date = date.today() - timedelta(days=auto_reprocess_last_n_days)
                    _reprocess = reprocess or (quarter_start > cutoff_date)
                    self.get_filing_list(year, quarter+1, reprocess=_reprocess)
        finally:
            settings.DEBUG = tmp_debug
            connection.close()
                
    def get_filing_list(self, year, quarter, reprocess=False):
        """
        Gets the list of filings and download locations for the given year and quarter.
        """
        url = 'ftp://ftp.sec.gov/edgar/full-index/%d/QTR%d/company.zip' % (year, quarter)
    
        # Download the data and save to a file
        if not os.path.isdir(DATA_DIR):
            os.makedirs(DATA_DIR)
        fn = os.path.join(DATA_DIR, 'company_%d_%d.zip' % (year, quarter))
    
        ifile, _ = IndexFile.objects.get_or_create(
            year=year, quarter=quarter, defaults=dict(filename=fn))
        if ifile.processed and not reprocess:
            return
        ifile.filename = fn
        
        if os.path.exists(fn) and reprocess:
            print('Deleting old file %s.' % fn)
            os.remove(fn)
        
        if not os.path.exists(fn):
            print('Downloading %s.' % (url,))
            try:
                compressed_data = urllib.urlopen(url).read()
            except IOError as e:
                print('Unable to download url: %s' % e)
                return
            fileout = open(fn, 'w')
            fileout.write(compressed_data)
            fileout.close()
            ifile.downloaded = timezone.now()
        
        if not ifile.downloaded:
            ifile.downloaded = timezone.now()
        ifile.save()
        transaction.commit()
        
        # Extract the compressed file
        print('Opening index file %s.' % (fn,))
        zipf = ZipFile(fn)
        zdata = zipf.read('company.idx')
        #zdata = removeNonAscii(zdata)
        
        # Parse the fixed-length fields
        bulk_companies = []
        bulk_indexes = []
        bulk_commit_freq = 1000
        status_secs = 3
        lines = zdata.split('\n')
        i = 0
        total = len(lines)
        IndexFile.objects.filter(id=ifile.id).update(total_rows=total)
        last_status = None
        prior_keys = set()
        #print('Found %i prior index keys.' % len(prior_keys)
        prior_ciks = set(Company.objects.all().values_list('cik', flat=True))
        print('Found %i prior ciks.' % len(prior_ciks))
        index_add_count = 0
        company_add_count = 0
        for r in lines[10:]: # Note, first 10 lines are useless headers.
            i += 1
            if not reprocess and ifile.processed_rows and i < ifile.processed_rows:
                continue
            if not last_status or ((datetime.now() - last_status).seconds >= status_secs):
                sys.stdout.write(
                    '\rProcessing record %i of %i (%.02f%%).' % (i, total, float(i)/total*100))
                sys.stdout.flush()
                last_status = datetime.now()
                IndexFile.objects.filter(id=ifile.id).update(processed_rows=i)
            dt = r[86:98].strip()
            if not dt:
                continue
            dt = date(*map(int, dt.split('-')))
            if r.strip() == '':
                continue
            name = r[0:62].strip()
            
            cik = int(r[74:86].strip())
            if cik not in prior_ciks:
                company_add_count += 1
                prior_ciks.add(cik)
                bulk_companies.append(Company(cik=cik, name=force_text(name, errors='replace')))
                
            filename = r[98:].strip()
            key = (cik, dt, filename)#, year, quarter)
            if key in prior_keys:
                continue
            prior_keys.add(key)
            if Index.objects.filter(company__cik=cik, date=dt, filename=filename).exists():
                continue
            index_add_count += 1
            bulk_indexes.append(Index(
                company_id=cik,
                form=r[62:74].strip(), # form type
                date=dt, # date filed
                year=year,
                quarter=quarter,
                filename=filename,
            ))
            if not len(bulk_indexes) % bulk_commit_freq:
                if len(bulk_companies):
                    Company.objects.bulk_create(bulk_companies)
                    bulk_companies = []
                Index.objects.bulk_create(bulk_indexes)
                bulk_indexes = []
                transaction.commit()
                
        if bulk_indexes:
            if len(bulk_companies):
                Company.objects.bulk_create(bulk_companies)
                bulk_companies = []
            Index.objects.bulk_create(bulk_indexes)
        IndexFile.objects.filter(id=ifile.id).update(processed=timezone.now())
        transaction.commit()
        
        print('\rProcessing record %i of %i (%.02f%%).' % (total, total, 100))
        print()
        print('%i new companies found.' % company_add_count)
        print('%i new indexes found.' % index_add_count)
        sys.stdout.flush()
        IndexFile.objects.filter(id=ifile.id).update(processed_rows=total)
        


from __future__ import print_function

import os
import datetime
from datetime import timedelta
import time
import socket
import threading
from functools import cmp_to_key
import warnings

import six

import django
from django.core.management import call_command
from django.core import mail
from django.test import TestCase
from django.test.client import Client
from django.utils import timezone
from django.contrib.auth.models import User
from django.conf import settings

from django_sec import models
from django_sec import utils
from django_sec import constants as c

warnings.simplefilter('error', RuntimeWarning)

socket.gethostname = lambda: 'localhost'

class Tests(TestCase):
    
    #fixtures = ['test_jobs.json']
    
    def setUp(self):
        pass
        
    def _test_example(self):
        
        # get a file from the index. it may or may not be present on our hard disk.
        # if it's not, it will be downloaded
        # the first time we try to access it, or you can call .download() explicitly
        filing = models.Index.objects\
            .filter(form='10-K', cik=1090872)\
            .order_by('-date')[0]
        
        print(filing.name)
        
        # initialize XBRL parser and populate an attribute called fields with a dict
        # of 50 common terms
        x = filing.xbrl()
        
        print(x.fields['FiscalYear'])
        
        print(x.fields)
        
        # fetch arbitrary XBRL tags representing eiter an Instant or a Duration in time
        print('Tax rate',
            x.GetFactValue('us-gaap:EffectiveIncomeTaxRateContinuingOperations', 'Duration'))
        
        if x.loadYear(1): 
            # Most 10-Ks have two or three previous years contained in them for the major values.
            # This call switches the contexts to the prior year (set it to 2 or 3 instead of 1 to
            # go back further) and reloads the fundamental concepts.
            # Any calls to GetFactValue will use that year's value from that point on.
                            
            print(x.fields['FiscalYear'])
        
            print(x.fields)
        
            print('Tax rate',
                x.GetFactValue('us-gaap:EffectiveIncomeTaxRateContinuingOperations', 'Duration'))
    
    def _test_sec_import_attrs(self):
        call_command('sec_import_attrs')
    
    def _test_sec_import_index(self):
        call_command('sec_import_index')
    
    def _test_sec_xbrl_to_csv(self):
        call_command('sec_xbrl_to_csv')
        
    def test_sec_mark_units(self):
        unit, _ = models.Unit.objects.get_or_create(name='U_iso4217USD')
        # In Django >= 1.9, you can't set a self-referential field during creation.
        unit.save()
        self.assertTrue(unit.true_unit)
        self.assertEqual(unit.true_unit, unit)
        
        call_command('sec_mark_units')



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Namespace'
        db.create_table(u'django_sec_namespace', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=500, db_index=True)),
        ))
        db.send_create_signal(u'django_sec', ['Namespace'])

        # Adding model 'Unit'
        db.create_table(u'django_sec_unit', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=200, db_index=True)),
            ('true_unit', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['django_sec.Unit'], null=True, on_delete=models.SET_NULL, blank=True)),
            ('master', self.gf('django.db.models.fields.BooleanField')(default=True)),
        ))
        db.send_create_signal(u'django_sec', ['Unit'])

        # Adding model 'Attribute'
        db.create_table(u'django_sec_attribute', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('namespace', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['django_sec.Namespace'])),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=500, db_index=True)),
            ('load', self.gf('django.db.models.fields.BooleanField')(default=False, db_index=True)),
            ('total_values', self.gf('django.db.models.fields.PositiveIntegerField')(null=True, blank=True)),
            ('total_values_fresh', self.gf('django.db.models.fields.BooleanField')(default=False)),
        ))
        db.send_create_signal(u'django_sec', ['Attribute'])

        # Adding unique constraint on 'Attribute', fields ['namespace', 'name']
        db.create_unique(u'django_sec_attribute', ['namespace_id', 'name'])

        # Adding index on 'Attribute', fields ['namespace', 'name']
        db.create_index(u'django_sec_attribute', ['namespace_id', 'name'])

        # Adding model 'AttributeValue'
        db.create_table(u'django_sec_attributevalue', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('company', self.gf('django.db.models.fields.related.ForeignKey')(related_name='attributes', to=orm['django_sec.Company'])),
            ('attribute', self.gf('django.db.models.fields.related.ForeignKey')(related_name='values', to=orm['django_sec.Attribute'])),
            ('value', self.gf('django.db.models.fields.DecimalField')(max_digits=20, decimal_places=6)),
            ('unit', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['django_sec.Unit'])),
            ('start_date', self.gf('django.db.models.fields.DateField')(db_index=True)),
            ('end_date', self.gf('django.db.models.fields.DateField')(null=True, blank=True)),
            ('filing_date', self.gf('django.db.models.fields.DateField')()),
        ))
        db.send_create_signal(u'django_sec', ['AttributeValue'])

        # Adding unique constraint on 'AttributeValue', fields ['company', 'attribute', 'start_date', 'end_date']
        db.create_unique(u'django_sec_attributevalue', ['company_id', 'attribute_id', 'start_date', 'end_date'])

        # Adding index on 'AttributeValue', fields ['company', 'attribute', 'start_date']
        db.create_index(u'django_sec_attributevalue', ['company_id', 'attribute_id', 'start_date'])

        # Adding model 'IndexFile'
        db.create_table(u'django_sec_indexfile', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('year', self.gf('django.db.models.fields.IntegerField')(db_index=True)),
            ('quarter', self.gf('django.db.models.fields.IntegerField')(db_index=True)),
            ('filename', self.gf('django.db.models.fields.CharField')(max_length=200)),
            ('total_rows', self.gf('django.db.models.fields.PositiveIntegerField')(null=True, blank=True)),
            ('processed_rows', self.gf('django.db.models.fields.PositiveIntegerField')(null=True, blank=True)),
            ('downloaded', self.gf('django.db.models.fields.DateTimeField')(null=True, blank=True)),
            ('processed', self.gf('django.db.models.fields.DateTimeField')(null=True, blank=True)),
        ))
        db.send_create_signal(u'django_sec', ['IndexFile'])

        # Adding unique constraint on 'IndexFile', fields ['year', 'quarter']
        db.create_unique(u'django_sec_indexfile', ['year', 'quarter'])

        # Adding model 'Company'
        db.create_table(u'django_sec_company', (
            ('cik', self.gf('django.db.models.fields.IntegerField')(primary_key=True, db_index=True)),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=100, db_index=True)),
            ('load', self.gf('django.db.models.fields.BooleanField')(default=False, db_index=True)),
        ))
        db.send_create_signal(u'django_sec', ['Company'])

        # Adding model 'Index'
        db.create_table(u'django_sec_index', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('company', self.gf('django.db.models.fields.related.ForeignKey')(related_name='filings', to=orm['django_sec.Company'])),
            ('form', self.gf('django.db.models.fields.CharField')(db_index=True, max_length=10, blank=True)),
            ('date', self.gf('django.db.models.fields.DateField')(db_index=True)),
            ('filename', self.gf('django.db.models.fields.CharField')(max_length=100, db_index=True)),
            ('year', self.gf('django.db.models.fields.IntegerField')(db_index=True)),
            ('quarter', self.gf('django.db.models.fields.IntegerField')(db_index=True)),
            ('attributes_loaded', self.gf('django.db.models.fields.BooleanField')(default=False, db_index=True)),
            ('valid', self.gf('django.db.models.fields.BooleanField')(default=True, db_index=True)),
            ('error', self.gf('django.db.models.fields.TextField')(null=True, blank=True)),
        ))
        db.send_create_signal(u'django_sec', ['Index'])

        # Adding unique constraint on 'Index', fields ['company', 'form', 'date', 'filename', 'year', 'quarter']
        db.create_unique(u'django_sec_index', ['company_id', 'form', 'date', 'filename', 'year', 'quarter'])

        # Adding index on 'Index', fields ['year', 'quarter']
        db.create_index(u'django_sec_index', ['year', 'quarter'])


    def backwards(self, orm):
        # Removing index on 'Index', fields ['year', 'quarter']
        db.delete_index(u'django_sec_index', ['year', 'quarter'])

        # Removing unique constraint on 'Index', fields ['company', 'form', 'date', 'filename', 'year', 'quarter']
        db.delete_unique(u'django_sec_index', ['company_id', 'form', 'date', 'filename', 'year', 'quarter'])

        # Removing unique constraint on 'IndexFile', fields ['year', 'quarter']
        db.delete_unique(u'django_sec_indexfile', ['year', 'quarter'])

        # Removing index on 'AttributeValue', fields ['company', 'attribute', 'start_date']
        db.delete_index(u'django_sec_attributevalue', ['company_id', 'attribute_id', 'start_date'])

        # Removing unique constraint on 'AttributeValue', fields ['company', 'attribute', 'start_date', 'end_date']
        db.delete_unique(u'django_sec_attributevalue', ['company_id', 'attribute_id', 'start_date', 'end_date'])

        # Removing index on 'Attribute', fields ['namespace', 'name']
        db.delete_index(u'django_sec_attribute', ['namespace_id', 'name'])

        # Removing unique constraint on 'Attribute', fields ['namespace', 'name']
        db.delete_unique(u'django_sec_attribute', ['namespace_id', 'name'])

        # Deleting model 'Namespace'
        db.delete_table(u'django_sec_namespace')

        # Deleting model 'Unit'
        db.delete_table(u'django_sec_unit')

        # Deleting model 'Attribute'
        db.delete_table(u'django_sec_attribute')

        # Deleting model 'AttributeValue'
        db.delete_table(u'django_sec_attributevalue')

        # Deleting model 'IndexFile'
        db.delete_table(u'django_sec_indexfile')

        # Deleting model 'Company'
        db.delete_table(u'django_sec_company')

        # Deleting model 'Index'
        db.delete_table(u'django_sec_index')


    models = {
        u'django_sec.attribute': {
            'Meta': {'unique_together': "(('namespace', 'name'),)", 'object_name': 'Attribute', 'index_together': "(('namespace', 'name'),)"},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'load': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '500', 'db_index': 'True'}),
            'namespace': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['django_sec.Namespace']"}),
            'total_values': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'total_values_fresh': ('django.db.models.fields.BooleanField', [], {'default': 'False'})
        },
        u'django_sec.attributevalue': {
            'Meta': {'ordering': "('-attribute__total_values', '-start_date', 'attribute__name')", 'unique_together': "(('company', 'attribute', 'start_date', 'end_date'),)", 'object_name': 'AttributeValue', 'index_together': "(('company', 'attribute', 'start_date'),)"},
            'attribute': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'values'", 'to': u"orm['django_sec.Attribute']"}),
            'company': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'attributes'", 'to': u"orm['django_sec.Company']"}),
            'end_date': ('django.db.models.fields.DateField', [], {'null': 'True', 'blank': 'True'}),
            'filing_date': ('django.db.models.fields.DateField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'start_date': ('django.db.models.fields.DateField', [], {'db_index': 'True'}),
            'unit': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['django_sec.Unit']"}),
            'value': ('django.db.models.fields.DecimalField', [], {'max_digits': '20', 'decimal_places': '6'})
        },
        u'django_sec.company': {
            'Meta': {'object_name': 'Company'},
            'cik': ('django.db.models.fields.IntegerField', [], {'primary_key': 'True', 'db_index': 'True'}),
            'load': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100', 'db_index': 'True'})
        },
        u'django_sec.index': {
            'Meta': {'unique_together': "(('company', 'form', 'date', 'filename', 'year', 'quarter'),)", 'object_name': 'Index', 'index_together': "(('year', 'quarter'),)"},
            'attributes_loaded': ('django.db.models.fields.BooleanField', [], {'default': 'False', 'db_index': 'True'}),
            'company': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'filings'", 'to': u"orm['django_sec.Company']"}),
            'date': ('django.db.models.fields.DateField', [], {'db_index': 'True'}),
            'error': ('django.db.models.fields.TextField', [], {'null': 'True', 'blank': 'True'}),
            'filename': ('django.db.models.fields.CharField', [], {'max_length': '100', 'db_index': 'True'}),
            'form': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '10', 'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'quarter': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'}),
            'valid': ('django.db.models.fields.BooleanField', [], {'default': 'True', 'db_index': 'True'}),
            'year': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'})
        },
        u'django_sec.indexfile': {
            'Meta': {'ordering': "('year', 'quarter')", 'unique_together': "(('year', 'quarter'),)", 'object_name': 'IndexFile'},
            'downloaded': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'filename': ('django.db.models.fields.CharField', [], {'max_length': '200'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'processed': ('django.db.models.fields.DateTimeField', [], {'null': 'True', 'blank': 'True'}),
            'processed_rows': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'quarter': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'}),
            'total_rows': ('django.db.models.fields.PositiveIntegerField', [], {'null': 'True', 'blank': 'True'}),
            'year': ('django.db.models.fields.IntegerField', [], {'db_index': 'True'})
        },
        u'django_sec.namespace': {
            'Meta': {'object_name': 'Namespace'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '500', 'db_index': 'True'})
        },
        u'django_sec.unit': {
            'Meta': {'object_name': 'Unit'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'master': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '200', 'db_index': 'True'}),
            'true_unit': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['django_sec.Unit']", 'null': 'True', 'on_delete': 'models.SET_NULL', 'blank': 'True'})
        }
    }

    complete_apps = ['django_sec']


from __future__ import print_function

import os
import csv

from django_sec import xbrl
from django_sec.models import *
from django.core.management.base import BaseCommand

import psycopg2

#TODO:remove? deprecated?
class Command(BaseCommand):
    help = "Put the 50+ common accounting terms from an arbitrary list of 10ks into a spreadsheet"
    
    def handle(self, csv_fn, **options):

        headers = [
            'EntityRegistrantName', 'EntityCentralIndexKey', 'EntityFilerCategory',
            'TradingSymbol', 'FiscalYear', 'FiscalPeriod', 'DocumentType', 'PeriodStartDate',
            'DocumentPeriodEndDate', 'Assets', 'CurrentAssets', 'NoncurrentAssets',
            'LiabilitiesAndEquity', 'Liabilities', 'CurrentLiabilities', 'NoncurrentLiabilities',
            'CommitmentsAndContingencies', 'TemporaryEquity', 'Equity',
            'EquityAttributableToParent', 'EquityAttributableToNoncontrollingInterest', 'Revenues',
            'CostOfRevenue', 'GrossProfit', 'OperatingExpenses', 'CostsAndExpenses',
            'OtherOperatingIncome', 'OperatingIncomeLoss', 'NonoperatingIncomeLoss',
            'InterestAndDebtExpense', 'NonoperatingIncomeLossPlusInterestAndDebtExpense',
            'IncomeBeforeEquityMethodInvestments', 'IncomeFromEquityMethodInvestments',
            'IncomeFromContinuingOperationsBeforeTax', 'IncomeTaxExpenseBenefit',
            'IncomeFromContinuingOperationsAfterTax', 'IncomeFromDiscontinuedOperations',
            'ExtraordaryItemsGainLoss', 'NetIncomeLoss', 'NetIncomeAttributableToParent',
            'NetIncomeAttributableToNoncontrollingInterest',
            'PreferredStockDividendsAndOtherAdjustments',
            'NetIncomeAvailableToCommonStockholdersBasic', 'ComprehensiveIncome',
            'OtherComprehensiveIncome', 'NetCashFlowsOperating',
            'NetCashFlowsOperatingContinuing', 'NetCashFlowsOperatingDiscontinued',
            'NetCashFlowsInvesting', 'NetCashFlowsInvestingContinuing',
            'NetCashFlowsInvestingDiscontinued', 'NetCashFlowsFinancing',
            'NetCashFlowsFinancingContinuing', 'NetCashFlowsFinancingDiscontinued',
            'NetCashFlowsContinuing', 'NetCashFlowsDiscontinued', 'ExchangeGainsLosses',
            'NetCashFlow', 'ComprehensiveIncomeAttributableToParent',
            'ComprehensiveIncomeAttributableToNoncontrollingInterest', 'SGR', 'ROA', 'ROE',
            'ROS', 'SECFilingPage', 'LinkToXBRLInstance',
        ]

        fout = csv.DictWriter(open(csv_fn, 'w'), headers)
        fout.writeheader()

        #this SQL is just a way of getting a list of particular CIKs I want
        conn = psycopg2.connect("dbname=recovery")
        cur = conn.cursor()
        cur.execute("SELECT cik, ticker FROM index WHERE cik is not null and use='1';")
        rows = cur.fetchall()
        for row in rows:
            cik = row[0]
            print('cik:', cik)
            for year in range(2011, 2014):
                latest = Index.objects.filter(
                    form='10-K',
                    cik=cik,
                    quarter__startswith=year
                ).order_by('-date')
                if len(latest):
                    latest = latest[0]
                    latest.download()
                    x = latest.xbrl()
                    if x is None:
                        print('no xbrl for ', cik, year)
                        continue
                                        
                    d = {}
                    for f in headers:
                        if f in x.fields.keys():
                            d[f] = x.fields[f]
                        else:
                            d[f] = ''
                    d['FiscalPeriod'] = x.fields['DocumentFiscalPeriodFocus']
                    d['FiscalYear'] = x.fields['DocumentFiscalYearFocus']
                    d['DocumentPeriodEndDate'] = x.fields['BalanceSheetDate']
                    d['PeriodStartDate'] = x.fields['IncomeStatementPeriodYTD']
                    d['SECFilingPage'] = latest.index_link()
                    d['LinkToXBRLInstance'] = latest.xbrl_link() 
                    
                    fout.writerow(d)




class FundamentantalAccountingConcepts:               

    def __init__(self, xbrl):               

        self.xbrl = xbrl        

        #self.xbrl.fields['Assets']
        
        #TODO:parse fixtures/normalized_fields.csv



from django.conf import settings
from django.contrib import admin
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse

try:
    from admin_steroids.queryset import ApproxCountQuerySet
except ImportError as e:
    ApproxCountQuerySet = None

from . import forms
from . import models

class NamespaceAdmin(admin.ModelAdmin):
    
    list_display = (
        'name',
    )
    
    search_fields = (
        'name',
    )

admin.site.register(
    models.Namespace,
    NamespaceAdmin)

class UnitAliasInlineAdmin(admin.TabularInline):
    
    model = models.Unit
    
    extra = 0
    max_num = 0
    can_delete = False
    
    list_display = (
        'name',
    )
    
    readonly_fields = (
        'name',
    )

class UnitAdmin(admin.ModelAdmin):
    
    form = forms.UnitChangeForm
    
    list_display = (
        'name',
        'true_unit',
        'master',
    )
    
    list_filter = (
        'master',
    )
    
    readonly_fields = (
        'master',
    )
    
    search_fields = (
        'name',
    )
    
    inlines = (
        UnitAliasInlineAdmin,
    )

admin.site.register(
    models.Unit,
    UnitAdmin)

class AttributeAdmin(admin.ModelAdmin):
    
    list_display = (
        'name',
        'namespace',
        'load',
        'total_values_fresh',
        'total_values',
    )
    
    list_filter = (
        'load',
        'total_values_fresh',
    )
    
    search_fields = (
        'name',
    )
    
    readonly_fields = (
        'total_values',
    )
    
    actions = (
        'enable_load',
        'disable_load',
        'refresh_total_values',
    )
    
    def queryset(self, *args, **kwargs):
        qs = super(AttributeAdmin, self).queryset(*args, **kwargs)
        if ApproxCountQuerySet:
            qs = qs._clone(klass=ApproxCountQuerySet)
        return qs
    
    def enable_load(self, request, queryset):
        models.Attribute.objects.filter(id__in=queryset).update(load=True)
        models.Index.objects.filter(attributes_loaded=True).update(attributes_loaded=False)
    enable_load.short_description = 'Enable value loading of selected %(verbose_name_plural)s'
    
    def disable_load(self, request, queryset):
        models.Attribute.objects.filter(id__in=queryset).update(load=False)
    disable_load.short_description = 'Disable value loading of selected %(verbose_name_plural)s'
    
    def refresh_total_values(self, request, queryset):
        queryset.update(total_values_fresh=False)
        models.Attribute.do_update()
    refresh_total_values.short_description = \
        'Refresh to total values count of selected %(verbose_name_plural)s'

admin.site.register(
    models.Attribute,
    AttributeAdmin)

class AttributeValueAdmin(admin.ModelAdmin):
    
    list_display = (
        'company_name',
        'attribute_name',
        'value',
        'true_unit',
        'start_date',
        'end_date',
        'filing_date',
        'attribute_total_values',
    )
    
    raw_id_fields = (
        'company',
        'attribute',
    )
    
    search_fields = (
        'company__name',
        'attribute__name',
    )
    
    readonly_fields = (
        'company_name',
        'attribute_name',
        'attribute_total_values',
        'true_unit',
    )
    
    exclude = (
        'unit',
    )
    
    def queryset(self, *args, **kwargs):
        qs = super(AttributeValueAdmin, self).queryset(*args, **kwargs)
        if ApproxCountQuerySet:
            qs = qs._clone(klass=ApproxCountQuerySet)
        return qs
    
    def true_unit(self, obj=None):
        if not obj:
            return ''
        return obj.unit.true_unit
    true_unit.short_description = 'unit'
    
    def company_name(self, obj=None):
        if not obj:
            return ''
        return obj.company.name
    
    def attribute_name(self, obj=None):
        if not obj:
            return ''
        return obj.attribute.name
    
    def attribute_total_values(self, obj=None):
        if not obj:
            return ''
        return obj.attribute.total_values
    attribute_total_values.admin_order_field = 'attribute__total_values'

admin.site.register(
    models.AttributeValue,
    AttributeValueAdmin)

class CompanyAdmin(admin.ModelAdmin):
    
    list_display = (
        'cik',
        'name',
        'min_date',
        'max_date',
        'load',
    )
    
    list_filter = (
        'load',
    )
    
    search_fields = (
        'cik',
        'name',
        '_ticker',
    )
    
    readonly_fields = (
        'cik',
        'name',
        'filings_link',
        'values_link',
        'min_date',
        'max_date',
    )
    
    actions = (
        'enable_load',
        'disable_load',
    )
    
    def lookup_allowed(self, key, value):
        return True
    
    def queryset(self, *args, **kwargs):
        qs = super(CompanyAdmin, self).queryset(*args, **kwargs)
        if ApproxCountQuerySet:
            qs = qs._clone(klass=ApproxCountQuerySet)
        return qs
    
    def enable_load(self, request, queryset):
        models.Company.objects.filter(cik__in=queryset).update(load=True)
        models.Index.objects\
            .filter(company__cik__in=queryset, attributes_loaded=True)\
            .update(attributes_loaded=False)
    enable_load.short_description = 'Enable attribute loading of selected %(verbose_name_plural)s'
    
    def disable_load(self, request, queryset):
        models.Company.objects.filter(cik__in=queryset).update(load=False)
    disable_load.short_description = 'Disable attribute loading of selected %(verbose_name_plural)s'
    
    def filings_link(self, obj=None):
        if not obj:
            return ''
        ct = ContentType.objects.get_for_model(models.Index)
        list_url_name = 'admin:%s_%s_changelist' % (ct.app_label, ct.model)
        url = reverse(list_url_name) + ('?company=%s' % obj.cik)
        count = obj.filings.all().count()
        return '<a href="%s" target="_blank" class="button">View %i</a>' % (url, count)
    filings_link.short_description = 'filings'
    filings_link.allow_tags = True
    
    def values_link(self, obj=None):
        if not obj:
            return ''
        ct = ContentType.objects.get_for_model(models.AttributeValue)
        list_url_name = 'admin:%s_%s_changelist' % (ct.app_label, ct.model)
        url = reverse(list_url_name) + ('?company=%s' % obj.cik)
        count = obj.attributes.all().count()
        return '<a href="%s" target="_blank" class="button">View %i</a>' % (url, count)
    values_link.short_description = 'attributes'
    values_link.allow_tags = True

admin.site.register(
    models.Company,
    CompanyAdmin)

class IndexFileAdmin(admin.ModelAdmin):
    
    list_display = (
        'year',
        'quarter',
        'total_rows',
        'processed_rows',
        'percent_processed',
        'downloaded',
        'processed',
    )
    
    readonly_fields = (
        'percent_processed',
    )
    
    actions = (
        'mark_unprocessed',
    )
    
    def queryset(self, *args, **kwargs):
        qs = super(IndexFileAdmin, self).queryset(*args, **kwargs)
        if ApproxCountQuerySet:
            qs = qs._clone(klass=ApproxCountQuerySet)
        return qs
    
    def mark_unprocessed(self, request, queryset):
        models.IndexFile.objects\
            .filter(id__in=queryset.values_list('id', flat=True))\
            .update(processed=None, processed_rows=0)
    mark_unprocessed.short_description = 'Mark selected %(verbose_name_plural)s as unprocessed'
    
    def percent_processed(self, obj=None):
        if not obj or not obj.total_rows or not obj.processed_rows:
            return ''
        return '%.02f%%' % (obj.processed_rows/float(obj.total_rows)*100,)

admin.site.register(
    models.IndexFile,
    IndexFileAdmin)
    
class IndexAdmin(admin.ModelAdmin):
    list_display = (
        'filename',
        'company',
        'cik',
        '_ticker',
        'form',
        'date',
        'quarter',
        'attributes_loaded',
        'valid',
    )
    
    search_fields = (
        'filename',
        'company__name',
    )
    
    list_filter = (
        'attributes_loaded',
        'valid',
        'year',
        'quarter',
        'form',
    )
    
    readonly_fields = (
        'cik',
        'xbrl_link',
    )
    
    actions = (
#        'enable',
#        'disable',
    )
    
    def queryset(self, *args, **kwargs):
        qs = super(IndexAdmin, self).queryset(*args, **kwargs)
        if ApproxCountQuerySet:
            qs = qs._clone(klass=ApproxCountQuerySet)
        return qs
    
    def cik(self, obj=None):
        if not obj:
            return ''
        return obj.company.cik
    cik.admin_order_field = 'company__cik'
    
    def get_readonly_fields(self, request, obj=None):
        exclude = []
        return [
            _.name for _ in self.model._meta.fields
            if _.name not in exclude
        ] + list(self.readonly_fields)
    
    def get_fieldsets(self, request, obj=None):
        readonly_fields = list(self.readonly_fields)
        exclude = readonly_fields + ['id']
        first = ['filename']
        fields = first + readonly_fields + [
            _.name for _ in self.model._meta.fields
            if _.name not in exclude and _.name not in first
        ]
        fieldsets = (
            (None, {
                'fields': fields,
            }),
        )
        return fieldsets
    
    def enable(self, request, queryset):
        for r in queryset.iterator():
            r.enabled = True
            r.save()
    enable.short_description = 'Enable selected %(verbose_name_plural)s'
    
    def disable(self, request, queryset):
        for r in queryset.iterator():
            r.enabled = False
            r.save()
    disable.short_description = 'Disable selected %(verbose_name_plural)s'
    
admin.site.register(
    models.Index,
    IndexAdmin)
